# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import sys
import typing
import logging
import contextlib
import click
import pyuavcan
from pyuavcan.presentation import Presentation, Subscriber
import yakut
from yakut.param.formatter import Formatter
from yakut.util import convert_transfer_metadata_to_builtin, construct_port_id_and_type


_M = typing.TypeVar("_M", bound=pyuavcan.dsdl.CompositeObject)


_logger = logging.getLogger(__name__)


@yakut.subcommand()
@click.argument("subject", type=str, nargs=-1)
@click.option(
    "--with-metadata/--no-metadata",
    "+M/-M",
    default=True,
    show_default=True,
    help="When enabled, each message object is prepended with an extra field named `_metadata_`.",
)
@click.option(
    "--count",
    "-N",
    type=int,
    metavar="CARDINAL",
    help=f"""
Exit automatically after this many messages (or synchronous message groups) have been received. No limit by default.
""",
)
@yakut.pass_purser
def subscribe(
    purser: yakut.Purser,
    subject: typing.Tuple[str, ...],
    with_metadata: bool,
    count: typing.Optional[int],
) -> None:
    """
    Subscribe to specified subjects and print messages into stdout.
    This command does not instantiate a local node and does not disturb the network in any way,
    so many instances can be cheaply executed concurrently.
    It is recommended to use anonymous transport (i.e., without a node-ID).

    The arguments are a list of message data type names prepended with the subject-ID;
    the subject-ID may be omitted if the data type defines a fixed one:

    \b
        [SUBJECT_ID.]TYPE_NAME.MAJOR.MINOR

    If multiple subjects are specified, a synchronous subscription will be used.
    It is useful for subscribing to a group of coupled subjects like lockstep sensor feeds,
    but it will not work for subjects that are temporally unrelated or published at different rates.

    Each object emitted into stdout is a key-value mapping where the number of elements equals the number
    of subjects the command is asked to subscribe to;
    the keys are subject-IDs and values are the received message objects.

    In data type names forward or backward slashes can be used instead of ".";
    version numbers can be also separated using underscores.
    This is done to allow the user to rely on filesystem autocompletion when typing the command.

    Examples:

    \b
        yakut sub 33.uavcan.si.unit.angle.Scalar.1.0 --no-metadata
    """
    _logger.debug("subject=%r, with_metadata=%r, count=%r", subject, with_metadata, count)
    if not subject:
        _logger.info("Nothing to do because no subjects are specified")
        return
    if count is not None and count <= 0:
        _logger.info("Nothing to do because count=%s", count)
        return

    count = count if count is not None else sys.maxsize
    formatter = purser.make_formatter()

    transport = purser.get_transport()
    if transport.local_node_id is not None:
        _logger.info("It is recommended to use an anonymous transport with this command.")

    with contextlib.closing(Presentation(transport)) as presentation:
        subscriber = _make_subscriber(subject, presentation)
        try:
            _run(subscriber, formatter, with_metadata=with_metadata, count=count)
        finally:
            if _logger.isEnabledFor(logging.INFO):
                _logger.info("%s", presentation.transport.sample_statistics())
                _logger.info("%s", subscriber.sample_statistics())


def _make_subscriber(subjects: typing.Sequence[str], presentation: Presentation) -> Subscriber[_M]:
    group = [construct_port_id_and_type(ds) for ds in subjects]
    assert len(group) > 0
    if len(group) == 1:
        ((subject_id, dtype),) = group
        return presentation.make_subscriber(dtype, subject_id)
    raise NotImplementedError(
        "Multi-subject subscription is not yet implemented. See https://github.com/UAVCAN/pyuavcan/issues/65"
    )


@yakut.asynchronous
async def _run(subscriber: Subscriber[_M], formatter: Formatter, with_metadata: bool, count: int) -> None:
    async for msg, transfer in subscriber:
        assert isinstance(transfer, pyuavcan.transport.TransferFrom)
        outer: typing.Dict[int, typing.Dict[str, typing.Any]] = {}

        bi: typing.Dict[str, typing.Any] = {}  # We use updates to ensure proper dict ordering: metadata before data
        if with_metadata:
            bi.update(convert_transfer_metadata_to_builtin(transfer))
        bi.update(pyuavcan.dsdl.to_builtin(msg))
        outer[subscriber.port_id] = bi

        print(formatter(outer))

        count -= 1
        if count <= 0:
            _logger.debug("Reached the specified message count, stopping")
            break



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import math
import typing
import asyncio
import logging
import contextlib
import click
import pyuavcan
import yakut
from yakut.yaml import YAMLLoader
from yakut.helpers import EnumParam


_MIN_SEND_TIMEOUT = 0.1
"""
With a slow garbage-collected language like Python, having a smaller timeout does not make practical sense.
This may be made configurable later.
"""

_logger = logging.getLogger(__name__)


def _validate_message_spec(
    ctx: click.Context,
    param: click.Parameter,
    value: typing.Tuple[str, ...],
) -> typing.List[typing.Tuple[str, str]]:
    if len(value) % 2 != 0:
        raise click.BadParameter(
            f"Message specifier shall have an even number of paired arguments (found {len(value)} arguments)",
            ctx=ctx,
            param=param,
        )
    return [(s, f) for s, f in (value[i : i + 2] for i in range(0, len(value), 2))]  # pylint: disable=R1721


@yakut.subcommand()
@click.argument(
    "message",
    type=str,
    callback=_validate_message_spec,
    metavar="SUBJECT FIELDS [SUBJECT FIELDS]...",
    nargs=-1,
)
@click.option(
    "--period",
    "-T",
    type=float,
    default=1.0,
    show_default=True,
    metavar="SECONDS",
    help=f"""
Message publication period.
All messages are published synchronously, so the period setting applies to all specified subjects.
The send timeout equals the period as long as it is not less than {_MIN_SEND_TIMEOUT} seconds.
""",
)
@click.option(
    "--count",
    "-N",
    type=int,
    default=1,
    show_default=True,
    metavar="CARDINAL",
    help=f"Number of publication cycles before exiting normally.",
)
@click.option(
    "--priority",
    "-P",
    default=pyuavcan.presentation.DEFAULT_PRIORITY,
    type=EnumParam(pyuavcan.transport.Priority),
    help=f"Priority of published message transfers. [default: {pyuavcan.presentation.DEFAULT_PRIORITY.name}]",
)
@yakut.pass_purser
def publish(
    purser: yakut.Purser,
    message: typing.Sequence[typing.Tuple[str, str]],
    period: float,
    count: int,
    priority: pyuavcan.transport.Priority,
) -> None:
    """
    Publish messages on the specified subjects.
    The local node will also publish heartbeat and respond to GetInfo, unless it is configured to be anonymous.

    The command accepts a list of space-separated pairs like:

    \b
        [SUBJECT_ID.]TYPE_NAME.MAJOR.MINOR  YAML_FIELDS

    The first element is a name like `uavcan.node.Heartbeat.1.0` prepended by the subject-ID.
    The subject-ID may be omitted if a fixed one is defined for the data type.

    The second element specifies the values of the message fields in YAML format (or JSON, which is a subset of YAML).
    Missing fields will be left at their default values;
    therefore, to publish a default-initialized message, the field specifier should be an empty dict: `{}`.
    For more info about the format see PyUAVCAN documentation on builtin-based representations.

    The number of such pairs can be arbitrary; all defined messages will be published synchronously.
    If no such pairs are specified, only the heartbeat will be published, unless the local node is anonymous.

    Forward or backward slashes can be used instead of ".";
    version numbers can be also separated using underscores.
    This is done to allow the user to rely on filesystem autocompletion when typing the command.

    Examples:

    \b
        yakut pub uavcan.diagnostic.Record.1.1 '{text: "Hello world!", severity: {value: 4}}' -N3 -T0.1 -P hi
        yakut pub 33.uavcan/si/unit/angle/Scalar_1_0 'radian: 2.31' uavcan.diagnostic.Record.1.1 'text: "2.31 rad"'
    """
    try:
        from pyuavcan.application import Node
    except ImportError as ex:
        from yakut.cmd.compile import make_usage_suggestion

        raise click.UsageError(make_usage_suggestion(ex.name))

    _logger.debug("period=%s, count=%s, priority=%s, message=%s", period, count, priority, message)
    assert all((isinstance(a, str) and isinstance(b, str)) for a, b in message)
    assert isinstance(period, float) and isinstance(count, int) and isinstance(priority, pyuavcan.transport.Priority)
    if period < 1e-9 or not math.isfinite(period):
        raise click.BadParameter("Period shall be a positive real number of seconds")
    if count <= 0:
        _logger.info("Nothing to do because count=%s", count)
        return

    send_timeout = max(_MIN_SEND_TIMEOUT, period)
    node = purser.get_node("publish", allow_anonymous=True)
    assert isinstance(node, Node)
    with contextlib.closing(node):
        publications = [
            Publication(
                subject_spec=subj,
                field_spec=fields,
                presentation=node.presentation,
                priority=priority,
                send_timeout=send_timeout,
            )
            for subj, fields in message
        ]
        if _logger.isEnabledFor(logging.INFO):
            _logger.info(
                "Ready to start publishing with period %.3fs, send timeout %.3fs, count %d, at %s:\n%s",
                period,
                send_timeout,
                count,
                priority,
                "\n".join(map(str, publications)) or "<nothing>",
            )
        try:  # Even if the publication set is empty, we still have to publish the heartbeat.
            _run(node, count=count, period=period, publications=publications)
        finally:
            if _logger.isEnabledFor(logging.INFO):
                _logger.info("%s", node.presentation.transport.sample_statistics())
                for s in node.presentation.transport.output_sessions:
                    ds = s.specifier.data_specifier
                    if isinstance(ds, pyuavcan.transport.MessageDataSpecifier):
                        _logger.info("Subject %d: %s", ds.subject_id, s.sample_statistics())


@yakut.asynchronous
async def _run(node: object, count: int, period: float, publications: typing.Sequence[Publication]) -> None:
    from pyuavcan.application import Node

    assert isinstance(node, Node)
    node.start()

    sleep_until = asyncio.get_event_loop().time()
    for c in range(count):
        out = await asyncio.gather(*[p.publish() for p in publications])
        assert len(out) == len(publications)
        assert all(isinstance(x, bool) for x in out)
        if not all(out):
            timed_out = [publications[idx] for idx, res in enumerate(out) if not res]
            _logger.error("The following publications have timed out:\n%s", "\n".join(map(str, timed_out)))

        sleep_until += period
        sleep_duration = sleep_until - asyncio.get_event_loop().time()
        _logger.info("Published group %6d of %6d; sleeping for %.3f seconds", c + 1, count, sleep_duration)
        await asyncio.sleep(sleep_duration)


class Publication:
    _YAML_LOADER = YAMLLoader()

    def __init__(
        self,
        subject_spec: str,
        field_spec: str,
        presentation: pyuavcan.presentation.Presentation,
        priority: pyuavcan.transport.Priority,
        send_timeout: float,
    ):
        from yakut.util import construct_port_id_and_type

        subject_id, dtype = construct_port_id_and_type(subject_spec)
        content = self._YAML_LOADER.load(field_spec)

        self._message = pyuavcan.dsdl.update_from_builtin(dtype(), content)
        self._publisher = presentation.make_publisher(dtype, subject_id)
        self._publisher.priority = priority
        self._publisher.send_timeout = send_timeout

    async def publish(self) -> bool:
        out = await self._publisher.publish(self._message)
        assert isinstance(out, bool)
        return out

    def __repr__(self) -> str:
        out = pyuavcan.util.repr_attributes(self, self._message, self._publisher)
        assert isinstance(out, str)
        return out



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import sys
import typing
import decimal
import logging
import contextlib
import click
import pyuavcan
import yakut
from yakut.helpers import EnumParam
from yakut.param.formatter import Formatter
from yakut.util import convert_transfer_metadata_to_builtin, construct_port_id_and_type


_S = typing.TypeVar("_S", bound=pyuavcan.dsdl.ServiceObject)


_logger = logging.getLogger(__name__)


def _validate_request_fields(ctx: click.Context, param: click.Parameter, value: str) -> typing.Any:
    from yakut.yaml import YAMLLoader

    try:
        fields = YAMLLoader().load(value)
    except Exception as ex:
        raise click.BadParameter(f"Could not parse the request object fields: {ex}", ctx=ctx, param=param)
    return fields


@yakut.subcommand()
@click.argument("server_node_id", metavar="SERVER_NODE_ID", type=int, required=True)
@click.argument("service", metavar="SERVICE", type=str, required=True)
@click.argument("request_fields", metavar="FIELDS", type=str, callback=_validate_request_fields, default="{}")
@click.option(
    "--timeout",
    "-T",
    type=float,
    default=pyuavcan.presentation.DEFAULT_SERVICE_REQUEST_TIMEOUT,
    show_default=True,
    metavar="SECONDS",
    help=f"Request timeout; how long to wait for the response before giving up.",
)
@click.option(
    "--priority",
    "-P",
    default=pyuavcan.presentation.DEFAULT_PRIORITY,
    type=EnumParam(pyuavcan.transport.Priority),
    help=f"Priority of the request transfer. [default: {pyuavcan.presentation.DEFAULT_PRIORITY.name}]",
)
@click.option(
    "--with-metadata/--no-metadata",
    "+M/-M",
    default=False,
    show_default=True,
    help="When enabled, the response object is prepended with an extra field named `_metadata_`.",
)
@yakut.pass_purser
def call(
    purser: yakut.Purser,
    server_node_id: int,
    service: str,
    request_fields: typing.Any,
    timeout: float,
    priority: pyuavcan.transport.Priority,
    with_metadata: bool,
) -> None:
    """
    Invoke an RPC-service using the specified request object and print the response.
    While waiting for the response, the local node will also publish heartbeat and respond to GetInfo.

    The first positional argument is the server node-ID.
    The second is the pair of service-ID (which can be omitted if a fixed one is defined for the type)
    and the data type name of the form:

    \b
        [SERVICE_ID.]FULL_SERVICE_TYPE_NAME.MAJOR.MINOR

    In the data type name, forward or backward slashes can be used instead of ".";
    version numbers can be also separated using underscores.
    This is done to allow the user to rely on filesystem autocompletion when typing the command.

    The third positional argument specifies the values of the request object fields in YAML format
    (or JSON, which is a subset of YAML).
    Missing fields will be left at their default values.
    If omitted, this argument defaults to an empty object: `{}`.
    For more info about the format see PyUAVCAN documentation on builtin-based representations.

    The output will be printed as a key-value mapping of one element where the key is the service-ID
    and the value is the received response object.

    Examples:

    \b
        yakut call 42 uavcan.node.GetInfo.1.0 +M -T3 -Pe
        yakut call 42 123.sirius_cyber_corp.PerformLinearLeastSquaresFit.1.0 'points: [{x: 10, y: 1}, {x: 20, y: 2}]'
    """
    try:
        from pyuavcan.application import Node
    except ImportError as ex:
        from yakut.cmd.compile import make_usage_suggestion

        raise click.UsageError(make_usage_suggestion(ex.name))

    _logger.debug(
        "server_node_id=%s, service=%r, request_fields=%r, timeout=%.6f, priority=%s, with_metadata=%s",
        server_node_id,
        service,
        request_fields,
        timeout,
        priority,
        with_metadata,
    )

    service_id, dtype = construct_port_id_and_type(service)
    if not issubclass(dtype, pyuavcan.dsdl.ServiceObject):
        raise TypeError(f"Expected a service type; got {dtype.__name__}")

    request = pyuavcan.dsdl.update_from_builtin(dtype.Request(), request_fields)
    _logger.info("Request object: %r", request)

    formatter = purser.make_formatter()
    node = purser.get_node("call", allow_anonymous=False)
    assert isinstance(node, Node) and callable(formatter)
    with contextlib.closing(node):
        client = node.presentation.make_client(dtype, service_id, server_node_id)
        client.response_timeout = timeout
        client.priority = priority
        node.start()
        _run(client, request, formatter, with_metadata=with_metadata)


@yakut.asynchronous
async def _run(
    client: pyuavcan.presentation.Client[_S],
    request: pyuavcan.dsdl.CompositeObject,
    formatter: Formatter,
    with_metadata: bool,
) -> None:
    request_ts_transport: typing.Optional[pyuavcan.transport.Timestamp] = None

    def on_transfer_feedback(fb: pyuavcan.transport.Feedback) -> None:
        nonlocal request_ts_transport
        request_ts_transport = fb.first_frame_transmission_timestamp

    client.output_transport_session.enable_feedback(on_transfer_feedback)

    request_ts_application = pyuavcan.transport.Timestamp.now()
    result = await client.call(request)
    response_ts_application = pyuavcan.transport.Timestamp.now()
    if result is None:
        click.secho(f"The request has timed out after {client.response_timeout:0.1f} seconds", err=True, fg="red")
        sys.exit(1)
    if not request_ts_transport:  # pragma: no cover
        request_ts_transport = request_ts_application
        _logger.warning(
            "The transport implementation is misbehaving: feedback was never emitted; "
            "falling back to software timestamping. "
            "Please submit a bug report. Involved instances: client=%r, result=%r",
            client,
            result,
        )
    response, transfer = result
    transport_duration = transfer.timestamp.monotonic - request_ts_transport.monotonic
    application_duration = response_ts_application.monotonic - request_ts_application.monotonic
    _logger.info(
        "Request duration [second]: "
        "transport layer: %.6f, application layer: %.6f, application layer overhead: %.6f",
        transport_duration,
        application_duration,
        application_duration - transport_duration,
    )

    bi: typing.Dict[str, typing.Any] = {}  # We use updates to ensure proper dict ordering: metadata before data
    if with_metadata:
        qnt = decimal.Decimal("0.000001")
        bi.update(
            convert_transfer_metadata_to_builtin(
                transfer,
                roundtrip_time={
                    "transport_layer": (transfer.timestamp.monotonic - request_ts_transport.monotonic).quantize(qnt),
                    "application_layer": application_duration.quantize(qnt),
                },
            )
        )
    bi.update(pyuavcan.dsdl.to_builtin(response))

    print(formatter({client.port_id: bi}))



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import sys
import random
import asyncio
import logging
import contextlib
import click
import pyuavcan
import yakut


_logger = logging.getLogger(__name__)


@yakut.subcommand()
@yakut.pass_purser
@yakut.asynchronous
async def accommodate(purser: yakut.Purser) -> None:
    """
    Automatically find a node-ID value that is not used by any other node that is currently online.

    This is a simpler alternative to plug-and-play node-ID allocation logic defined in Specification.
    Unlike the solution presented there, this alternative is less deterministic and less robust;
    it is fundamentally unsafe and it should not be used in production.
    Instead, it is intended for use in R&D and testing applications,
    either directly by humans or from automation scripts.

    The operating principle is extremely simple and can be viewed as a simplification of the
    node-ID claiming procedure defined in J1939:
    listen to Heartbeat messages for a short while,
    build the list of node-ID values that are currently in use,
    and then randomly pick a node-ID from the unused ones.
    The listening duration is determined heuristically at run time;
    for most use cases it is unlikely to exceed three seconds.
    """
    try:
        import uavcan.node
    except (ImportError, AttributeError):
        from yakut.cmd.compile import make_usage_suggestion

        raise click.UsageError(make_usage_suggestion("uavcan")) from None

    transport = purser.get_transport()
    node_id_set_cardinality = transport.protocol_parameters.max_nodes
    if node_id_set_cardinality > 2 ** 24:
        # Special case: for very large sets just pick a random number.
        # Very large sets are only possible with test transports such as loopback so it's acceptable.
        # If necessary, later we could develop a more robust solution.
        click.echo(random.randint(0, node_id_set_cardinality - 1))
        return

    candidates = set(range(node_id_set_cardinality))
    try:
        candidates.remove(transport.local_node_id)  # Allow non-anonymous transports for consistency.
    except LookupError:
        pass

    if node_id_set_cardinality > 1000:
        # Special case: some transports with large NID cardinality may have difficulties supporting a node-ID of zero
        # depending on the configuration of the underlying hardware and software.
        # This is not a problem of UAVCAN but of the platform itself.
        # For example, a UDP/IP transport over IPv4 with a node-ID of zero would map to an IP address with trailing
        # zeros which may be the address of the subnet, which is likely to cause all sorts of complications.
        _logger.debug("Removing the zero node-ID from the set of available values to avoid platform-specific issues")
        candidates.remove(0)

    pres = pyuavcan.presentation.Presentation(transport)
    with contextlib.closing(pres):
        deadline = asyncio.get_event_loop().time() + uavcan.node.Heartbeat_1_0.MAX_PUBLICATION_PERIOD * 2.0
        sub = pres.make_subscriber_with_fixed_subject_id(uavcan.node.Heartbeat_1_0)
        while asyncio.get_event_loop().time() <= deadline:
            result = await sub.receive(deadline)
            if result is None:
                break
            msg, transfer = result
            assert isinstance(transfer, pyuavcan.transport.TransferFrom)
            _logger.debug("Received %r via %r", msg, transfer)
            if transfer.source_node_id is None:
                _logger.warning(
                    "FYI, the network contains an anonymous node which is publishing Heartbeat. "
                    "Please contact the vendor and inform them that this behavior is non-compliant. "
                    "The offending heartbeat message is: %r, transfer: %r",
                    msg,
                    transfer,
                )
            else:
                try:
                    candidates.remove(int(transfer.source_node_id))
                except LookupError:
                    pass
                else:
                    # If at least one node is in the Initialization state, the network might be starting,
                    # so we need to listen longer to minimize the chance of collision.
                    multiplier = 3.0 if msg.mode.value == uavcan.node.Mode_1_0.INITIALIZATION else 1.0
                    advancement = uavcan.node.Heartbeat_1_0.MAX_PUBLICATION_PERIOD * multiplier
                    _logger.debug(
                        "Deadline advanced by %.1f s; %d candidates left of %d possible",
                        advancement,
                        len(candidates),
                        node_id_set_cardinality,
                    )
                    deadline = max(deadline, asyncio.get_event_loop().time() + advancement)

    if not candidates:
        click.secho(f"All {node_id_set_cardinality} of the available node-ID values are occupied.", err=True, fg="red")
        sys.exit(1)
    else:
        pick = random.choice(list(candidates))
        _logger.info(
            "The set of unoccupied node-ID values contains %d elements out of %d possible; the chosen value is %d",
            len(candidates),
            node_id_set_cardinality,
            pick,
        )
        click.echo(pick)



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import typing
import click


Formatter = typing.Callable[[typing.Any], str]
FormatterFactory = typing.Callable[[], Formatter]


def formatter_factory_option(f: typing.Callable[..., typing.Any]) -> typing.Callable[..., typing.Any]:
    def validate(ctx: click.Context, param: object, value: str) -> FormatterFactory:
        _ = ctx, param
        try:
            return _FORMATTERS[value.upper()]
        except LookupError:
            raise click.BadParameter(f"Invalid format name: {value!r}") from None

    choices = list(_FORMATTERS.keys())
    default = choices[0]
    doc = f"""
The format of data printed into stdout.
This option is only relevant for commands that generate structured outputs, like pub or call; other commands ignore it.

The final representation of the output data is constructed from an intermediate "builtin-based" representation,
which is a simplified form that is stripped of the detailed DSDL type information, like JSON.
For more info please read the PyUAVCAN documentation on builtin-based representations.

YAML separates objects with `---`.

JSON and TSV (tab separated values) keep exactly one object per line.

TSV is intended for use with third-party software
such as computer algebra systems or spreadsheet processors.
"""
    f = click.option(
        "--format",
        "-F",
        "formatter_factory",
        envvar="YAKUT_FORMAT",
        type=click.Choice(choices, case_sensitive=False),
        callback=validate,
        default=default,
        show_default=True,
        help=doc,
    )(f)
    return f


def _make_yaml_formatter() -> Formatter:
    from yakut.yaml import YAMLDumper

    dumper = YAMLDumper(explicit_start=True)
    return dumper.dumps


def _make_json_formatter() -> Formatter:
    # We prefer simplejson over the standard json because the native json lacks important capabilities:
    #  - simplejson preserves dict ordering, which is very important for UX.
    #  - simplejson supports Decimal.
    import simplejson as json

    return lambda data: json.dumps(data, ensure_ascii=False, separators=(",", ":"))


def _make_tsv_formatter() -> Formatter:
    # TODO print into a TSV (tab separated values, like CSV with tabs instead of commas).
    # The TSV format should place one scalar per column for ease of parsing by third-party software.
    # Variable-length entities such as arrays should expand into the maximum possible number of columns?
    # Unions should be represented by adjacent groups of columns where only one such group contains values?
    # We may need to obtain the full type information here in order to build the final representation.
    # Sounds complex. Search for better ways later. We just need a straightforward way of dumping data into a
    # standard tabular format for later processing using third-party software.
    raise NotImplementedError("Sorry, the TSV formatter is not yet implemented")


_FORMATTERS = {
    "YAML": _make_yaml_formatter,
    "JSON": _make_json_formatter,
    "TSV": _make_tsv_formatter,
}


def _unittest_formatter() -> None:
    obj = {
        2345: {
            "abc": {
                "def": [
                    123,
                    456,
                ],
            },
            "ghi": 789,
        }
    }
    assert (
        _FORMATTERS["YAML"]()(obj)
        == """---
2345:
  abc:
    def:
    - 123
    - 456
  ghi: 789
"""
    )
    assert _FORMATTERS["JSON"]()(obj) == '{"2345":{"abc":{"def":[123,456]},"ghi":789}}'



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import sys
import typing
import logging
import click
import yakut


_logger = logging.getLogger(__name__)


@yakut.subcommand()
@click.argument("name", default="")
def documentation(name: str) -> None:
    """
    Show transport usage documentation from PyUAVCAN.
    Transports whose dependencies are not installed will not be shown.

    If the argument NAME is provided, the documentation will be shown only for entities whose name contains
    the specified string (case-insensitive), like "udp".

    Full documentation is available at https://pyuavcan.readthedocs.io
    """
    import pydoc
    import pyuavcan

    fill_width = click.get_terminal_size()[0] - 1

    # noinspection PyTypeChecker
    pyuavcan.util.import_submodules(pyuavcan.transport, error_handler=_handle_import_error)
    transport_base = pyuavcan.transport.Transport
    texts: typing.List[str] = []
    for cls in pyuavcan.util.iter_descendants(transport_base):
        if not cls.__name__.startswith("_") and cls is not transport_base:
            public_module = cls.__module__.split("._")[0]
            public_name = public_module + "." + cls.__name__
            if name.lower() in public_name.lower():
                texts.append(
                    "\n".join(
                        [
                            "-" * fill_width,
                            public_name.center(fill_width, " "),
                            "-" * fill_width,
                            cls.__doc__,
                            pydoc.text.document(cls.__init__),
                            "",
                        ]
                    )
                )
    if texts:
        click.echo_via_pager(texts)
    else:
        click.secho(f"There are no entries that match {name!r}", err=True, fg="red")
        sys.exit(1)


def _handle_import_error(name: str, ex: ImportError) -> None:
    _logger.info("Transport module %r is not available because: %r", name, ex)



# Copyright (c) 2020 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

if __name__ == "__main__":
    from yakut import main

    main(auto_envvar_prefix="YAKUT")  # pylint: disable=no-value-for-parameter,unexpected-keyword-arg



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import typing
import inspect
import logging
import pyuavcan
from pyuavcan.transport import Transport as Transport
import click
from yakut.paths import OUTPUT_TRANSFER_ID_MAP_DIR, OUTPUT_TRANSFER_ID_MAP_MAX_AGE


TransportFactory = typing.Callable[[], typing.Optional[Transport]]
"""
The result is None if no transport configuration was provided when invoking the command.
"""

_logger = logging.getLogger(__name__)


def transport_factory_option(f: typing.Callable[..., typing.Any]) -> typing.Callable[..., typing.Any]:
    def validate(ctx: click.Context, param: object, value: typing.Optional[str]) -> TransportFactory:
        _ = ctx, param
        _logger.debug("Transport expression: %r", value)

        def factory() -> typing.Optional[Transport]:
            result: typing.Optional[Transport] = None
            if value:
                try:
                    result = construct_transport(value)
                except Exception as ex:
                    raise click.BadParameter(f"Could not initialize transport {value!r}: {ex!r}") from ex
            _logger.info("Expression %r yields %r", value, result)
            return result

        return factory

    doc = f"""
Specify the UAVCAN network interface to use.
This option is only relevant for commands that access the network, like pub/sub/call/etc.; other commands ignore it.

The value is a (Python) expression that yields a transport instance or a sequence thereof upon evaluation.
In the latter case, the multiple transports will be joined under the same redundant transport instance,
which may be heterogeneous (e.g., UDP+Serial).

The node-ID for the local node is to be configured here as well, because per the UAVCAN architecture,
this is a transport-layer property.

To see supported transports and how they should be initialized, run `yakut doc`.
Also, read the PyUAVCAN documentation at https://pyuavcan.readthedocs.io.

The transport expression does not need to explicitly reference the `pyuavcan.transport` module
because its contents are wildcard-imported for convenience.
Further, when specifying a transport class, the suffix `Transport` may be omitted;
e.g., `UDPTransport` and `UDP` are equivalent.

Examples showcasing loopback, CAN, and heterogeneous UDP+Serial:

\b
    Loopback(None)
    CAN(can.media.socketcan.SocketCANMedia('vcan0',64),42)
    UDP('127.42.0.123',anonymous=True),Serial("/dev/ttyUSB0",None)

To avoid issues caused by resetting the transfer-ID counters between invocations,
the tool stores the output transfer-ID map on disk keyed by the node-ID.
On this computer, the path is `{OUTPUT_TRANSFER_ID_MAP_DIR}`.
The map files can be removed to reset all transfer-ID counters to zero.
Files that are more than {OUTPUT_TRANSFER_ID_MAP_MAX_AGE} seconds old are not used.
"""
    f = click.option(
        "--transport",
        "-i",
        "transport_factory",
        envvar="YAKUT_TRANSPORT",
        type=str,
        metavar="EXPRESSION",
        callback=validate,
        help=doc,
    )(f)
    return f


def construct_transport(expression: str) -> Transport:
    context = _make_evaluation_context()
    trs = _evaluate_transport_expr(expression, context)
    _logger.debug("Transport expression evaluation result: %r", trs)
    if len(trs) == 1:
        return trs[0]  # Non-redundant transport
    if len(trs) > 1:
        from pyuavcan.transport.redundant import RedundantTransport

        rt = RedundantTransport()
        for t in trs:
            rt.attach_inferior(t)
        assert rt.inferiors == trs
        return rt
    raise ValueError("No transports specified")


def _evaluate_transport_expr(expression: str, context: typing.Dict[str, typing.Any]) -> typing.List[Transport]:
    out = eval(expression, context)  # pylint: disable=eval-used
    if isinstance(out, Transport):
        return [out]
    if isinstance(out, (list, tuple)) and all(isinstance(x, Transport) for x in out):
        return list(out)
    raise ValueError(
        f"The expression {expression!r} yields an instance of {type(out).__name__!r}. "
        f"Expected an instance of pyuavcan.transport.Transport or a list thereof."
    )


def _make_evaluation_context() -> typing.Dict[str, typing.Any]:
    import os

    def handle_import_error(parent_module_name: str, ex: ImportError) -> None:
        try:
            tr = parent_module_name.split(".")[2]
        except LookupError:
            tr = parent_module_name
        _logger.debug("Transport %r is not available due to the missing dependency %r", tr, ex.name)

    # This import is super slow, so we do it as late as possible.
    # Doing this when generating command-line arguments would be disastrous for performance.
    # noinspection PyTypeChecker
    pyuavcan.util.import_submodules(pyuavcan.transport, error_handler=handle_import_error)

    # Populate the context with all references that may be useful for the transport expression.
    context: typing.Dict[str, typing.Any] = {
        "pyuavcan": pyuavcan,
        "os": os,
    }

    # Expose pre-imported transport modules for convenience.
    for name, module in inspect.getmembers(pyuavcan.transport, inspect.ismodule):
        if not name.startswith("_"):
            context[name] = module

    # Pre-import transport classes.
    for cls in pyuavcan.util.iter_descendants(Transport):
        if not cls.__name__.startswith("_") and cls is not Transport:
            name = cls.__name__.rpartition(Transport.__name__)[0]
            assert name
            context[name] = cls
            context[cls.__name__] = cls

    _logger.debug("Transport expression evaluation context (on the next line):\n%r", context)
    return context



# Copyright (c) 2020 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import sys
import pyuavcan.util

# BY CONVENTION, the COMMAND and the MODULE it is defined in should be NAMED IDENTICALLY.
pyuavcan.util.import_submodules(sys.modules[__name__])



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import os
import re
import time
import atexit
import pickle
import typing
import logging
import pathlib
import dataclasses
import click
import pyuavcan
from pyuavcan.transport import Transport, OutputSessionSpecifier, Priority
from pyuavcan.presentation import OutgoingTransferIDCounter
import yakut
from yakut.paths import OUTPUT_TRANSFER_ID_MAP_DIR, OUTPUT_TRANSFER_ID_MAP_MAX_AGE
from yakut.helpers import EnumParam


_logger = logging.getLogger(__name__)


@dataclasses.dataclass()
class NodeFactory:
    """
    Constructs a node instance. The instance must be start()ed by the caller afterwards.
    """

    heartbeat_vssc: typing.Optional[int] = None
    heartbeat_period: typing.Optional[float] = None
    heartbeat_priority: typing.Optional[Priority] = None

    node_info: typing.Dict[str, typing.Any] = dataclasses.field(
        default_factory=lambda: {
            "protocol_version": {
                "major": pyuavcan.UAVCAN_SPECIFICATION_VERSION[0],
                "minor": pyuavcan.UAVCAN_SPECIFICATION_VERSION[1],
            },
            "software_version": {
                "major": yakut.__version_info__[0],
                "minor": yakut.__version_info__[1],
            },
        }
    )

    def __call__(self, transport: Transport, name_suffix: str, allow_anonymous: bool) -> object:
        """
        We use ``object`` for return type instead of Node because the Node class requires generated code
        to be generated.
        """
        _logger.debug("Constructing node using %r with %r and name %r", self, transport, name_suffix)
        if not re.match(r"[a-z][a-z0-9_]*[a-z0-9]", name_suffix):  # pragma: no cover
            raise ValueError(f"Internal error: Poorly chosen node name suffix: {name_suffix!r}")

        try:
            from pyuavcan import application
        except ImportError as ex:
            from yakut.cmd.compile import make_usage_suggestion

            raise click.UsageError(make_usage_suggestion(ex.name))

        try:
            node_info = pyuavcan.dsdl.update_from_builtin(application.NodeInfo(), self.node_info)
        except (ValueError, TypeError) as ex:
            raise click.UsageError(f"Node info fields are not valid: {ex}") from ex
        if len(node_info.name) == 0:
            node_info.name = f"org.uavcan.yakut.{name_suffix}"
        _logger.debug("Node info: %r", node_info)

        presentation = pyuavcan.presentation.Presentation(transport)
        node = application.Node(presentation, info=node_info)
        try:
            # Configure the heartbeat publisher.
            try:
                if self.heartbeat_period is not None:
                    node.heartbeat_publisher.period = self.heartbeat_period
                if self.heartbeat_priority is not None:
                    node.heartbeat_publisher.priority = self.heartbeat_priority
                if self.heartbeat_vssc is not None:
                    node.heartbeat_publisher.vendor_specific_status_code = self.heartbeat_vssc
                else:
                    node.heartbeat_publisher.vendor_specific_status_code = os.getpid() % 100
            except ValueError as ex:
                raise click.UsageError(f"Invalid heartbeat parameters: {ex}") from ex
            _logger.debug(
                "Node heartbeat: %s, period: %s, priority: %s",
                node.heartbeat_publisher.make_message(),
                node.heartbeat_publisher.period,
                node.heartbeat_publisher.priority,
            )

            # Check the node-ID configuration.
            if not allow_anonymous and node.presentation.transport.local_node_id is None:
                raise click.UsageError(
                    "The specified transport is configured in anonymous mode, which cannot be used with this command."
                )

            # Configure the transfer-ID map.
            # Register save on exit even if we're anonymous because the local node-ID may be provided later.
            _register_output_transfer_id_map_save_at_exit(node.presentation)
            # Restore if we have a node-ID. If we don't, no restoration will take place even if the node-ID is
            # provided later. This behavior is acceptable for CLI; a regular UAVCAN application will not need
            # to deal with saving/restoration at all since this use case is specific to CLI only.
            path = _get_output_transfer_id_map_path(node.presentation.transport)
            tid_map_restored = False
            if path is not None:
                tid_map = _restore_output_transfer_id_map(path)
                if tid_map:
                    _logger.debug("Restored output TID map from %s: %r", path, tid_map)
                    presentation.output_transfer_id_map.update(tid_map)
                    tid_map_restored = True
            if not tid_map_restored:
                _logger.debug("Could not restore output TID map from %s", path)

            _logger.info("Constructed node: %s", node)
            return node
        except Exception:
            node.close()
            raise


def node_factory_option(f: typing.Callable[..., typing.Any]) -> typing.Callable[..., typing.Any]:
    factory = NodeFactory()

    def validate_heartbeat_vssc(ctx: click.Context, param: click.Parameter, value: typing.Optional[str]) -> None:
        _ = ctx, param
        if value is not None:
            factory.heartbeat_vssc = int(value)

    def validate_heartbeat_period(ctx: click.Context, param: click.Parameter, value: typing.Optional[str]) -> None:
        _ = ctx, param
        if value is not None:
            factory.heartbeat_period = float(value)

    def validate_heartbeat_priority(
        ctx: click.Context, param: click.Parameter, value: typing.Optional[Priority]
    ) -> None:
        _ = ctx, param
        if value is not None:
            factory.heartbeat_priority = value

    def validate(ctx: click.Context, param: click.Parameter, value: str) -> NodeFactory:
        from yakut.yaml import YAMLLoader

        fields = YAMLLoader().load(value)
        if not isinstance(fields, dict):
            raise click.BadParameter(f"Expected a dict, got {type(fields).__name__}", ctx=ctx, param=param)

        factory.node_info.update(fields)
        return factory

    f = click.option(
        "--heartbeat-vssc",
        "--vssc",
        type=int,
        expose_value=False,
        callback=validate_heartbeat_vssc,
        help=f"""
The vendor-specific status code (VSSC) of the local node. The default is (PID % 100) of the current process.
""",
    )(f)
    f = click.option(
        "--heartbeat-period",
        type=float,
        metavar="SECONDS",
        callback=validate_heartbeat_period,
        expose_value=False,
        help=f"Heartbeat publication interval.",
    )(f)
    f = click.option(
        "--heartbeat-priority",
        type=EnumParam(Priority),
        callback=validate_heartbeat_priority,
        expose_value=False,
        help=f"Heartbeat publication priority.",
    )(f)
    f = click.option(
        "--node-info",
        "node_factory",
        envvar="YAKUT_NODE_INFO",
        default="{}",
        metavar="YAML",
        type=str,
        callback=validate,
        help=f"""
Override the default values of the uavcan.node.GetInfo response returned by the local node.
The defaults are (the default node name depends on the subcommand):

\b
{factory.node_info}
""",
    )(f)
    return f


def _restore_output_transfer_id_map(
    file_path: pathlib.Path,
) -> typing.Dict[OutputSessionSpecifier, OutgoingTransferIDCounter]:
    try:
        with open(str(file_path), "rb") as f:
            tid_map = pickle.load(f)
    except Exception as ex:  # pylint: disable=broad-except
        _logger.info("Output TID map: Could not restore from file %s: %s: %s", file_path, type(ex).__name__, ex)
        return {}
    mtime_abs_diff = abs(file_path.stat().st_mtime - time.time())
    if mtime_abs_diff > OUTPUT_TRANSFER_ID_MAP_MAX_AGE:
        _logger.debug("Output TID map: File %s is valid but too old: mtime age diff %.0f s", file_path, mtime_abs_diff)
        return {}
    if isinstance(tid_map, dict) and all(isinstance(v, OutgoingTransferIDCounter) for v in tid_map.values()):
        return tid_map
    _logger.warning("Output TID map file %s contains invalid data of type %s", file_path, type(tid_map).__name__)
    return {}


def _register_output_transfer_id_map_save_at_exit(presentation: pyuavcan.presentation.Presentation) -> None:
    # We MUST sample the configuration early because if this is a redundant transport it may reset its
    # configuration (local node-ID) back to default after close().
    path = _get_output_transfer_id_map_path(presentation.transport)
    _logger.debug("Output TID map file for %s: %s", presentation.transport, path)

    def do_save_at_exit() -> None:
        if path is not None:
            tmp = f"{path}.{os.getpid()}.{time.time_ns()}.tmp"
            _logger.debug("Output TID map save: %s --> %s", tmp, path)
            with open(tmp, "wb") as f:
                pickle.dump(presentation.output_transfer_id_map, f)
            # We use replace for compatibility reasons. On POSIX, a call to rename() will be made, which is
            # guaranteed to be atomic. On Windows this may fall back to non-atomic copy, which is still
            # acceptable for us here. If the file ends up being damaged, we'll simply ignore it at next startup.
            os.replace(tmp, str(path))
            try:
                os.unlink(tmp)
            except OSError:
                pass

    atexit.register(do_save_at_exit)


def _get_output_transfer_id_map_path(transport: Transport) -> typing.Optional[pathlib.Path]:
    if transport.local_node_id is not None:
        path = OUTPUT_TRANSFER_ID_MAP_DIR / str(transport.local_node_id)
        path.parent.mkdir(parents=True, exist_ok=True)
        return path
    return None


def _unittest_output_tid_file_path() -> None:
    from pyuavcan.transport.redundant import RedundantTransport
    from pyuavcan.transport.loopback import LoopbackTransport

    def once(tr: Transport) -> typing.Optional[pathlib.Path]:
        return _get_output_transfer_id_map_path(tr)

    assert once(LoopbackTransport(None)) is None
    assert once(LoopbackTransport(123)) == OUTPUT_TRANSFER_ID_MAP_DIR / "123"

    red = RedundantTransport()
    assert once(red) is None
    red.attach_inferior(LoopbackTransport(4000))
    red.attach_inferior(LoopbackTransport(4000))
    assert once(red) == OUTPUT_TRANSFER_ID_MAP_DIR / "4000"

    red = RedundantTransport()
    red.attach_inferior(LoopbackTransport(None))
    red.attach_inferior(LoopbackTransport(None))
    assert once(red) is None
